#!/usr/bin/env python3

import os
import sys

import matplotlib
import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import torch
import yaml
from natsort import natsorted
from sklearn.metrics import (
    balanced_accuracy_score,
    classification_report,
    confusion_matrix,
    matthews_corrcoef,
    roc_auc_score,
)
from collections import defaultdict
from pprint import pprint


class ClassificationReporter(object):
    """Saves the following metrics under sdir.
       - Balanced Accuracy
       - MCC
       - Confusion Matrix
       - Classification Report
       - ROC AUC score / curve
       - PRE-REC AUC score / curve

    Manual adding example:
        ##############################
        ## fig
        ##############################
        fig, ax = plt.subplots()
        ax.plot(np.random.rand(10))
        reporter.add(
            "manu_figs",
            fig,
            {
                "dirname": "manu_fig_dir/",
                "ext": ".png",
            },
        )
        ##############################
        ## DataFrame
        ##############################
        df = pd.DataFrame(np.random.rand(5, 3))
        reporter.add("manu_dfs", df, {"fname": "manu_dfs.csv", "method": "mean"})
        ##############################
        ## scalar
        ##############################
        scalar = random.random()
        reporter.add(
            "manu_scalars",
            scalar,
            {"fname": "manu_scalars.csv", "column_name": "manu_column_name"},
        )
    """

    def __init__(self, sdir):
        self.sdir = sdir
        # self.ts = mngs.general.TimeStamper()

        self.folds_dict = defaultdict(list)

        # print("\n{}\n".format(mngs.general.gen_timestamp()))
        # self.ts("\nReporter has been initialized.\n")

    def add(
        self,
        obj_name,
        obj,
    ):
        assert isinstance(obj_name, str)
        self.folds_dict[obj_name].append(obj)

    def calc_metrics(
        self,
        true_class,
        pred_class,
        pred_proba,
        labels=None,
        i_fold=None,
        show=True,
    ):
        """Calculates ACC, Confusion Matrix, Classification Report, and ROC-AUC score."""
        self.labels = labels

        true_class, pred_class, pred_proba = (
            mngs.general.torch_to_arr(true_class),
            mngs.general.torch_to_arr(pred_class),
            mngs.general.torch_to_arr(pred_proba),
        )

        ####################
        ## Scalar metrices
        ####################
        acc = (true_class.reshape(-1) == pred_class.reshape(-1)).mean()
        balanced_acc = balanced_accuracy_score(
            true_class.reshape(-1), pred_class.reshape(-1)
        )
        mcc = float(matthews_corrcoef(true_class.reshape(-1), pred_class.reshape(-1)))

        if show:
            print(f"\nACC in fold#{i_fold} was {acc:.3f}\n")
            print(f"\nBalanced ACC in fold#{i_fold} was {balanced_acc:.3f}\n")
            print(f"\nMCC in fold#{i_fold} was {mcc:.3f}\n")

        ####################
        ## Confusion Matrix
        ####################
        conf_mat = pd.DataFrame(
            data=confusion_matrix(true_class, pred_class), columns=labels
        ).set_index(pd.Series(list(labels)))
        if show:
            print(f"\nConfusion Matrix in fold#{i_fold}: \n")
            pprint(conf_mat)
            print()

        ####################
        ## Classification Report
        ####################
        clf_report = pd.DataFrame(
            classification_report(
                true_class,
                pred_class,
                target_names=labels,
                output_dict=True,
            )
        )
        ## ACC to bACC
        clf_report["accuracy"] = balanced_acc
        clf_report = clf_report.rename(columns={"accuracy": "balanced accuracy"})
        clf_report = clf_report.round(3)
        ## Renames 'support' to 'sample size'
        clf_report["index"] = clf_report.index
        clf_report.loc["support", "index"] = "sample size"
        clf_report.set_index("index", drop=True, inplace=True)
        clf_report.index.name = None
        if show:
            print(f"\nClassification Report for fold#{i_fold}:\n")
            pprint(clf_report)
            print()

        ####################
        ## ROC-AUC score
        ####################
        mngs.plt.configure_mpl(
            plt,
            figsize=(7, 7),
            labelsize=8,
            fontsize=7,
            legendfontsize=6,
            tick_size=0.8,
            tick_width=0.2,
        )

        fig_roc, metrics_roc_auc_dict = mngs.ml.plt.roc_auc(
            plt,
            true_class,
            pred_proba,
            labels,
        )
        plt.close()

        ####################
        ## PRE-REC AUC score
        ####################
        fig_pre_rec, metrics_pre_rec_auc_dict = mngs.ml.plt.pre_rec_auc(
            plt, true_class, pred_proba, labels
        )
        plt.close()

        ####################
        ## to the buffer
        ####################
        self.folds_dict["mcc"].append(mcc)
        self.folds_dict["balanced_acc"].append(balanced_acc)
        self.folds_dict["conf_mat"].append(conf_mat)
        self.folds_dict["clf_report"].append(clf_report)
        self.folds_dict["roc_auc_micro"].append(
            metrics_roc_auc_dict["roc_auc"]["micro"]
        )
        self.folds_dict["roc_auc_macro"].append(
            metrics_roc_auc_dict["roc_auc"]["macro"]
        )
        self.folds_dict["roc_auc_fig"].append(fig_roc)

        self.folds_dict["pre_rec_auc_micro"].append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["micro"]
        )
        self.folds_dict["pre_rec_auc_macro"].append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["macro"]
        )
        self.folds_dict["pre_rec_auc_fig"].append(fig_pre_rec)

    def summarize(
        self,
        n_round=3,
        show=False,
    ):
        """
        1) Take mean and std of scalars/pd.Dataframes for folds.
        2) Replace self.folds_dict with the summarized DataFrames.
        """
        _n_folds = [len(self.folds_dict[k]) for k in self.folds_dict.keys()]
        assert len(np.unique(_n_folds)) == 1
        self.n_folds = n_folds = _n_folds[0]
        self.cv_index = [f"{n_folds}-folds_CV_mean", f"{n_folds}-fold_CV_std"] + [
            f"fold#{i_fold}" for i_fold in range(n_folds)
        ]

        for k in self.folds_dict.keys():
            ## listed scalars
            if mngs.general.is_listed_X(self.folds_dict[k], [float, int]):
                mm = np.mean(self.folds_dict[k])
                ss = np.std(self.folds_dict[k], ddof=1)

                sr = pd.DataFrame(
                    data=[mm, ss] + self.folds_dict[k], index=self.cv_index, columns=[k]
                )
                self.folds_dict[k] = sr.round(n_round)

                if show:
                    print("\n----------------------------------------\n")
                    print(f"\n{self.folds_dict[k].iloc[:2]}\n")
                    print("\n----------------------------------------\n")

            ## listed pd.DataFrames
            elif mngs.general.is_listed_X(self.folds_dict[k], pd.DataFrame):
                zero_df_for_mm = 0 * self.folds_dict[k][0].copy()
                zero_df_for_ss = 0 * self.folds_dict[k][0].copy()

                mm = (zero_df_for_mm + np.stack(self.folds_dict[k]).mean(axis=0)).round(
                    n_round
                )
                ss = (
                    zero_df_for_ss + np.stack(self.folds_dict[k]).std(axis=0, ddof=1)
                ).round(n_round)

                self.folds_dict[k] = [mm, ss] + [
                    df_fold.round(n_round) for df_fold in self.folds_dict[k]
                ]

                if show:
                    print("\n----------------------------------------\n")
                    print(f"\n{k}\n")
                    print(f"\n{n_folds}-fold-CV mean:\n")
                    pprint(self.folds_dict[k][0])
                    print()
                    print(f"\n{n_folds}-fold-CV std.:\n")
                    pprint(self.folds_dict[k][1])
                    print()
                    print("\n----------------------------------------\n")

            ## listed figures
            elif mngs.general.is_listed_X(self.folds_dict[k], matplotlib.figure.Figure):
                pass

            else:
                print(f"{k} was not summarized")
                print(type(self.folds_dict[k][0]))

    def save(
        self,
        meta_dict=None,
    ):
        """
        1) Saves the content of self.folds_dict.
        2) Plots the colormap of confusion matrices and saves them.
        3) Saves passed meta_dict under self.sdir

        Example:
            meta_df_1 = pd.DataFrame(data=np.random.rand(3,3))
            meta_dict_1 = {"a": 0}
            meta_dict_2 = {"b": 0}
            meta_dict = {"meta_1.csv": meta_df_1,
                         "meta_1.yaml": meta_dict_1,
                         "meta_2.yaml": meta_dict_1,
            }

        """
        if meta_dict is not None:
            for k, v in meta_dict.items():
                mngs.general.save(v, self.sdir + k)

        for k in self.folds_dict.keys():

            ## pd.Series / pd.DataFrame
            if isinstance(self.folds_dict[k], pd.Series) or isinstance(
                self.folds_dict[k], pd.DataFrame
            ):
                mngs.general.save(self.folds_dict[k], self.sdir + f"{k}.csv")

            ## listed pd.DataFrame
            elif mngs.general.is_listed_X(self.folds_dict[k], pd.DataFrame):
                mngs.general.save(
                    self.folds_dict[k],
                    self.sdir + f"{k}.csv",
                    indi_suffix=self.cv_index,
                )

            ## listed figures
            elif mngs.general.is_listed_X(self.folds_dict[k], matplotlib.figure.Figure):
                for i_fold, fig in enumerate(self.folds_dict[k]):
                    mngs.general.save(
                        self.folds_dict[k][i_fold], self.sdir + f"{k}/fold#{i_fold}.png"
                    )

            else:
                print(f"{k} was not saved")
                print(type(self.folds_dict[k]))

        # self._plot_and_save_conf_mats()  # fixme

    def plot_and_save_conf_mats(
        self, plt, extend_ratio=1.0, colorbar=True, sci_notation_kwargs=None
    ):  # koko
        def _plot_conf_mat(
            plt, cm_df, title, extend_ratio=1.0, colorbar=True, sci_notation_kwargs=None
        ):
            labels = list(cm_df.columns)
            fig_conf_mat = mngs.ml.plt.confusion_matrix(
                plt,
                cm_df,
                labels=labels,
                title=title,
                extend_ratio=extend_ratio,
                colorbar=colorbar,
            )

            if sci_notation_kwargs is not None:
                fig_conf_mat.axes[-1] = mngs.plt.ax_scientific_notation(
                    fig_conf_mat.axes[-1], **sci_notation_kwargs
                )
            return fig_conf_mat

        ########################################
        ## Prepares confmats dfs
        ########################################
        ## Drops mean and std for the folds
        try:
            conf_mats = self.folds_dict["conf_mat"][-self.n_folds :]
        except:
            conf_mats = self.folds_dict["conf_mat"]

        ## Prepaires conf_mat_overall_sum
        conf_mat_zero = 0 * conf_mats[0].copy()  # get the table format
        conf_mat_overall_sum = conf_mat_zero + np.stack(conf_mats).sum(axis=0)

        ########################################
        ## Plots & Saves
        ########################################
        # each fold's conf
        for i_fold, cm in enumerate(conf_mats):
            title = f"Test fold#{i_fold}"
            fig_conf_mat_fold = _plot_conf_mat(
                plt,
                cm,
                title,
                extend_ratio=extend_ratio,
                colorbar=colorbar,
                sci_notation_kwargs=sci_notation_kwargs,
            )
            mngs.general.save(
                fig_conf_mat_fold, self.sdir + f"conf_mat_figs/fold#{i_fold}.png"
            )
            plt.close()

        ## overall_sum conf_mat
        title = f"{self.n_folds}-CV overall sum"
        fig_conf_mat_overall_sum = _plot_conf_mat(
            plt,
            conf_mat_overall_sum,
            title,
            extend_ratio=extend_ratio,
            colorbar=colorbar,
            sci_notation_kwargs=sci_notation_kwargs,
        )
        mngs.general.save(
            fig_conf_mat_overall_sum,
            self.sdir + f"conf_mat_figs/k-fold_cv_overall-sum.png",
        )
        plt.close()


if __name__ == "__main__":
    import random

    import mngs
    import numpy as np
    from catboost import CatBoostClassifier, Pool
    from sklearn.datasets import load_digits
    from sklearn.model_selection import StratifiedKFold

    import sys

    ################################################################################
    ## Sets tee
    ################################################################################
    sdir = mngs.general.path.mk_spath("")  # "/tmp/sdir/"
    sys.stdout, sys.stderr = mngs.general.tee(sys, sdir)

    ################################################################################
    ## Fixes seeds
    ################################################################################
    mngs.general.fix_seeds(np=np)

    ## Loads
    mnist = load_digits()
    X, T = mnist.data, mnist.target
    labels = mnist.target_names.astype(str)

    ## Main
    skf = StratifiedKFold(n_splits=5, shuffle=True)
    reporter = ClassificationReporter(sdir)
    for i_fold, (indi_tra, indi_tes) in enumerate(skf.split(X, T)):
        X_tra, T_tra = X[indi_tra], T[indi_tra]
        X_tes, T_tes = X[indi_tes], T[indi_tes]

        clf = CatBoostClassifier(verbose=False)

        clf.fit(X_tra, T_tra, verbose=False)

        ## Prediction
        pred_proba_tes = clf.predict_proba(X_tes)
        pred_cls_tes = np.argmax(pred_proba_tes, axis=1)

        ##############################
        ## Manually adds objects to reporter to save
        ##############################
        ## Figure
        fig, ax = plt.subplots()
        ax.plot(np.arange(10))
        reporter.add("manu_figs", fig)

        ## DataFrame
        df = pd.DataFrame(np.random.rand(5, 3))
        reporter.add("manu_dfs", df)

        ## Scalar
        scalar = random.random()
        reporter.add(
            "manu_scalars",
            scalar,
        )

        ########################################
        ## Metrics
        ########################################
        reporter.calc_metrics(
            T_tes,
            pred_cls_tes,
            pred_proba_tes,
            labels=labels,
            i_fold=i_fold,
        )

    reporter.summarize(show=True)
    reporter.save()

    ## Configures mpl for confmats
    mngs.plt.configure_mpl(
        plt,
        figsize=(8, 8),
        # labelsize=8,
        # fontsize=6,
        # legendfontsize=6,
        figscale=2,
        tick_size=0.8,
        tick_width=0.2,
    )

    sci_notation_kwargs = dict(
        order=1,
        fformat="%1.0d",
        scilimits=(-3, 3),
        x=False,
        y=True,
    )  # "%3.1f"

    # sci_notation_kwargs = None
    reporter.plot_and_save_conf_mats(
        plt, extend_ratio=1.0, sci_notation_kwargs=sci_notation_kwargs
    )

    ## EOF



#!/usr/bin/env python3

import time


class TimeStamper:
    def __init__(self, format="simple"):
        self.format = format
        self.time = time
        self.id = -1
        self.start = time.time()
        self.prev = self.start

    def __call__(self, comment=""):
        now = self.time.time()
        from_start = now - self.start

        self.from_start_hhmmss = self.time.strftime(
            "%H:%M:%S", self.time.gmtime(from_start)
        )
        from_prev = now - self.prev

        self.from_prev_hhmmss = self.time.strftime(
            "%H:%M:%S", self.time.gmtime(from_prev)
        )

        self.id += 1
        self.prev = now

        if self.format == "simple":
            return "ID:{} | {} {} | ".format(self.id, self.from_start_hhmmss, comment)

        else:
            return "Time (id:{}): tot {}, prev {} [hh:mm:ss]: {}\n".format(
                self.id, self.from_start_hhmmss, self.from_prev_hhmmss, comment
            )

    def get(self):
        return self.id, self.from_start_hhmmss, self.from_prev_hhmmss, comment



#!/usr/bin/env python3

from . import plt
from .LearningCurveLogger import LearningCurveLogger
from .merge_labels import merge_labels
from .ClassificationReporter import ClassificationReporter
from .under_sample import under_sample



#!/usr/bin/env python3

import numpy as np

import pandas as pd


################################################################################
## Pandas
################################################################################
def col_to_last(df, col):
    df_orig = df.copy()
    cols = list(df_orig.columns)
    cols_remain = pop_keys_from_keys_list(cols, col)
    out = pd.concat((df_orig[cols_remain], df_orig[col]), axis=1)
    return out


def col_to_top(df, col):
    df_orig = df.copy()
    cols = list(df_orig.columns)
    cols_remain = pop_keys_from_keys_list(cols, col)
    out = pd.concat((df_orig[col], df_orig[cols_remain]), axis=1)
    return out


class IDAllocator(object):
    """
    Example1:
        patterns = np.array([3, 2, 1, 2, 50, 20])
        alc = IDAllocator(patterns)
        input = np.array([2, 20, 3, 1])
        IDs = alc(input)
        print(IDs) # [1, 3, 2, 0]

    Example2:
        patterns = np.array(['a', 'b', 'c', 'zzz'])
        alc = IDAllocator(patterns)
        input = np.array(['c', 'a', 'zzz', 'b'])
        IDs = alc(input)
        print(IDs) # [2, 0, 3, 1]
    """

    def __init__(self, patterns):
        patterns_uq = np.unique(patterns)  # natural sorting is executed.
        new_IDs = np.arange(len(patterns_uq))
        self.correspondence_table = pd.DataFrame(
            {
                "Original": patterns_uq,
                "new_ID": new_IDs,
            }
        ).set_index("Original")

    def __call__(self, x):
        allocated = np.array(self.correspondence_table.loc[x]).squeeze()
        return allocated


def force_dataframe(permutable_dict, filler=""):
    ## Deep copy
    permutable_dict = permutable_dict.copy()

    ## Get the lengths
    max_len = 0
    for k, v in permutable_dict.items():
        max_len = max(max_len, len(v))

    ## Replace with 0 filled list
    for k, v in permutable_dict.items():
        permutable_dict[k] = list(v) + (max_len - len(v)) * [filler]

    ## Puts them into a DataFrame
    out_df = pd.DataFrame(permutable_dict)

    return out_df



#!/usr/bin/env python3

import numpy as np
import torch


def torch_to_arr(x):
    is_arr = isinstance(x, (np.ndarray, np.generic))
    if is_arr:  # when x is np.array
        return x
    if torch.is_tensor(x):  # when x is torch.tensor
        return x.detach().cpu().numpy()


def cvt_multi2single_model_state_dict(state_dict_multi):
    from collections import OrderedDict

    new_state_dict = OrderedDict()
    for k, v in state_dict_multi.items():
        name = k
        if name.startswith("module."):
            name = name[7:]  # remove 'module.' of dataparallel
        new_state_dict[name] = v
    return new_state_dict
    



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from scipy import fftpack


def calc_fft_powers(signals_2d, samp_rate):
    """
    Example:
        sig_len = 1024
        n_sigs = 32
        signals_2d = np.random.rand(n_sigs, sig_len)
        samp_rate = 256
        fft_df = calc_fft_powers(signals_2d, samp_rate)
    """
    fft_powers = np.abs(fftpack.fft(signals_2d))
    fft_freqs = np.fft.fftfreq(signals_2d.shape[-1], d=1.0 / samp_rate)
    mask = fft_freqs >= 0
    fft_powers, fft_freqs = fft_powers[:, mask], np.round(fft_freqs[mask], 1)
    fft_df = pd.DataFrame(data=fft_powers, columns=fft_freqs.astype(str))
    return fft_df



#!/usr/bin/env python3

from .cuda_collect_env import main as cuda_collect_env
from .debug import paste, reload
from .latex import add_hat_in_the_latex_style, to_the_latex_style
from .load import get_data_path_from_a_package, load
from .mat2py import *
from .misc import (
    decapitalize,
    connect_nums,
    connect_strs,
    does_exist,
    fmt_size,
    grep,
    isclose,
    listed_dict,
    pop_keys,
    search,
    squeeze_spaces,
    take_the_closest,
)
from .pandas import col_to_last, col_to_top, force_dataframe
from .path import *
from .repro import *
from .save import is_listed_X, save, save_listed_dfs_as_csv, save_listed_scalars_as_csv
from .TimeStamper import *
from .torch import *



#!/usr/bin/env python3

################################################################################
## Debugging
################################################################################
def paste():
    from IPython import embed

    embed()


def reload(module_or_func):
    import importlib
    import sys

    try:
        importlib.reload(module_orfunc)
    except:
        pass

    try:
        importlib.reload(sys.modules[module_orfunc.__module__])
    except:
        pass


# def reload_func(func):
#     import importlib
#     import sys

#     # importlib.reload(sys.modules['foo'])
#     importlib.reload(sys.modules[func.__module__])
#     # from foo import bar
#     # from sys.modules[func.__module__] import func
#     return func


# def reload_module(module):
#     import importlib

#     module = importlib.reload(module)
#     return module


# def modify_func(func):
#     module = sys.modules[func.__module__]


# def modify_func(fun):
#     modulename = fun.__module__
#     module = __import__(modulename)


# def modify_func(fun):
#     import inspect

#     module = inspect.getmodule(fun)



#!/usr/bin/env python3

from scipy.signal import butter, sosfilt, sosfreqz


def bandpassfilter(data, lo_hz, hi_hz, fs, order=5):
    def _mk_butter_bandpass(order=5):
        nyq = 0.5 * fs
        low, high = lo_hz / nyq, hi_hz / nyq
        sos = butter(order, [low, high], analog=False, btype="band", output="sos")
        return sos

    def _butter_bandpass_filter(data):
        sos = _mk_butter_bandpass()
        y = sosfilt(sos, data)
        return y

    sos = _mk_butter_bandpass(order=order)
    y = _butter_bandpass_filter(data)

    return y



#!/usr/bin/env python3


def load(lpath, show=False, **kwargs):
    import pickle

    import h5py
    import numpy as np
    import pandas as pd
    import torch
    import yaml

    # csv
    if lpath.endswith(".csv"):
        obj = pd.read_csv(lpath, **kwargs)
    # numpy
    if lpath.endswith(".npy"):
        obj = np.load(lpath)
    # pkl
    if lpath.endswith(".pkl"):
        with open(lpath, "rb") as l:
            obj = pickle.load(l)
    # joblib
    if lpath.endswith(".joblib"):
        with open(lpath, "rb") as l:
            obj = joblib.load(l)
    # hdf5
    if lpath.endswith(".hdf5"):
        obj = {}
        with h5py.File(fpath, "r") as hf:
            for name in name_list:
                obj_tmp = hf[name][:]
                obj[name] = obj_tmp
    # png
    if lpath.endswith(".png"):
        pass
    # tiff
    if lpath.endswith(".tiff") or lpath.endswith(".tif"):
        pass
    # yaml
    if lpath.endswith(".yaml"):
        obj = {}
        with open(lpath) as f:
            obj_tmp = yaml.safe_load(f)
            obj.update(obj_tmp)
    # txt
    if lpath.endswith(".txt"):
        f = open(lpath, "r")
        obj = [l.strip("\n\r") for l in f]
        f.close()
    # pth
    if lpath.endswith(".pth"):
        # return model.load_state_dict(torch.load(lpath))
        obj = torch.load(lpath)

    if lpath.endswith(".mat"):
        import pymatreader

        obj = pymatreader.read_mat(lpath)
    # xml
    if lpath.endswith("xml"):
        from ._xml2dict import xml2dict

        obj = xml2dict(lpath)

    if show:
        print("\nLoaded: {}\n".format(lpath))

    return obj


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def get_data_path_from_a_package(package_str, resource):
    import importlib
    import os
    import sys

    spec = importlib.util.find_spec(package_str)
    data_dir = os.path.join(spec.origin.split("src")[0], "data")
    resource_path = os.path.join(data_dir, resource)
    return resource_path



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from obspy.signal.tf_misfit import cwt


def wavelet(wave, samp_rate, f_min=100, f_max=None, plot=False):
    dt = 1.0 / samp_rate
    npts = len(wave)
    t = np.linspace(0, dt * npts, npts)
    if f_min == None:
        f_min = 0.1
    if f_max == None:
        f_max = int(samp_rate / 2)
    scalogram = cwt(wave, dt, 8, f_min, f_max)

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        x, y = np.meshgrid(
            t, np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0])
        )

        ax.pcolormesh(x, y, np.abs(scalogram), cmap=obspy_sequential)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Frequency [Hz]")
        ax.set_yscale("log")
        ax.set_ylim(f_min, f_max)
        plt.show()

    Hz = pd.Series(np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]))
    df = pd.DataFrame(np.abs(scalogram))
    df["Hz"] = Hz
    df.set_index("Hz", inplace=True)

    return df



#!/usr/bin/env python3

import csv

import pandas as pd
import mngs
import numpy as np


def save(obj, sfname_or_spath, makedirs=True, show=True, **kwargs):
    """
    Example
      save(arr, 'data.npy')
      save(df, 'df.csv')
      save(serializable, 'serializable.pkl')
    """
    import inspect
    import os
    import pickle

    import h5py
    import numpy as np
    import pandas as pd
    import torch
    import yaml

    spath, sfname = None, None

    if "/" in sfname_or_spath:
        spath = sfname_or_spath
    else:
        sfname = sfname_or_spath

    if (spath is None) and (sfname is not None):
        ## for ipython
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = "/tmp/fake.py"

        ## spath
        fpath = __file__
        fdir, fname, _ = mngs.general.split_fpath(fpath)
        sdir = fdir + fname + "/"
        spath = sdir + sfname
        # spath = mk_spath(sfname, makedirs=True)

    ## Make directory
    if makedirs:
        sdir = os.path.dirname(spath)
        os.makedirs(sdir, exist_ok=True)

    ## Saves
    # csv
    if spath.endswith(".csv"):
        if isinstance(obj, pd.DataFrame):  # DataFrame
            obj.to_csv(spath)
        if is_listed_X(obj, [int, float]):  # listed scalars
            save_listed_scalars_as_csv(
                obj,
                spath,
                **kwargs,
            )
        if is_listed_X(obj, pd.DataFrame):  # listed DataFrame
            # save_listed_dfs_as_csv(obj, spath, indi_suffix=None, overwrite=False)
            save_listed_dfs_as_csv(obj, spath, **kwargs)
    # numpy
    if spath.endswith(".npy"):
        np.save(spath, obj)
    # pkl
    if spath.endswith(".pkl"):
        with open(spath, "wb") as s:  # 'w'
            pickle.dump(obj, s)
    # joblib
    if spath.endswith(".joblib"):
        with open(spath, "wb") as s:  # 'w'
            joblib.dump(obj, s, compress=3)
    # png
    if spath.endswith(".png"):
        try:
            obj.savefig(spath)
        except:
            obj.figure.savefig(spath)
        del obj
    # tiff
    if spath.endswith(".tiff") or spath.endswith(".tif"):
        obj.savefig(spath, dpi=300, format="tiff")  # obj is matplotlib.pyplot object
        del obj
    # mp4
    if spath.endswith(".mp4"):
        mk_mp4(obj, spath)  # obj is matplotlib.pyplot.figure object
        del obj
    # yaml
    if spath.endswith(".yaml"):
        with open(spath, "w") as f:
            yaml.dump(obj, f)
    # hdf5
    if spath.endswith(".hdf5"):
        name_list, obj_list = []
        for k, v in obj.items():
            name_list.append(k)
            obj_list.append(v)
        with h5py.File(spath, "w") as hf:
            for (name, obj) in zip(name_list, obj_list):
                hf.create_dataset(name, data=obj)
    # pth
    if spath.endswith(".pth"):
        # torch.save(obj.state_dict(), spath)
        torch.save(obj, spath)

    if show:
        print("\nSaved to: {s}\n".format(s=spath))


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def is_listed_X(obj, types):
    """
    Example:
        obj = [3, 2, 1, 5]
        is_listed_X(obj,
    """
    import numpy as np

    try:
        conditions = []
        condition_list = isinstance(obj, list)

        if not (isinstance(types, list) or isinstance(types, tuple)):
            types = [types]

        _conditions_susp = []
        for typ in types:
            _conditions_susp.append(
                (np.array([isinstance(o, typ) for o in obj]) == True).all()
            )

        condition_susp = np.any(_conditions_susp)

        is_listed_X = np.all([condition_list, condition_susp])
        return is_listed_X

    except:
        return False


def save_listed_scalars_as_csv(
    listed_scalars,
    spath_csv,
    column_name="_",
    indi_suffix=None,
    round=3,
    overwrite=False,
    show=False,
):
    """Puts to df and save it as csv"""
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)
    indi_suffix = np.arange(len(listed_scalars)) if indi_suffix is None else indi_suffix
    df = pd.DataFrame(
        {"{}".format(column_name): listed_scalars}, index=indi_suffix
    ).round(round)
    df.to_csv(spath_csv)
    if show:
        print("\nSaved to: {}\n".format(spath_csv))


def save_listed_dfs_as_csv(
    listed_dfs,
    spath_csv,
    indi_suffix=None,
    overwrite=False,
    show=False,
):
    """listed_dfs:
        [df1, df2, df3, ..., dfN]. They will be written vertically in the order.

    spath_csv:
        /hoge/fuga/foo.csv

    indi_suffix:
        At the left top cell on the output csv file, '{}'.format(indi_suffix[i])
        will be added, where i is the index of the df.On the other hand,
        when indi_suffix=None is passed, only '{}'.format(i) will be added.
    """
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)

    indi_suffix = np.arange(len(listed_dfs)) if indi_suffix is None else indi_suffix
    for i, df in enumerate(listed_dfs):
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            i_suffix = indi_suffix[i]
            f_writer.writerow(["{}".format(indi_suffix[i])])
        df.to_csv(spath_csv, mode="a", index=True, header=True)
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            f_writer.writerow([""])
    if show:
        print("Saved to: {}".format(spath_csv))


def mk_mp4(fig, spath_mp4):
    from matplotlib import animation

    axes = fig.get_axes()

    def init():
        return (fig,)

    def animate(i):
        for ax in axes:
            ax.view_init(elev=10.0, azim=i)
        return (fig,)

    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=360, interval=20, blit=True
    )

    writermp4 = animation.FFMpegWriter(fps=60, extra_args=["-vcodec", "libx264"])
    anim.save(spath_mp4, writer=writermp4)
    print("\nSaving to: {}\n".format(spath_mp4))


def mv_to_tmp(fpath, L=2):
    import os
    from shutil import move

    try:
        tgt_fname = connect_strs_with_hyphens(fpath.split("/")[-L:])
        tgt_fpath = "/tmp/{}".format(tgt_fname)
        move(fpath, tgt_fpath)
        print("Moved to: {}".format(tgt_fpath))
    except:
        pass


# def save(obj, sfname_or_spath, makedirs=True):
#     """
#     Example
#       save(arr, 'data.npy')
#       save(df, 'df.csv')
#       save(serializable, 'serializable.pkl')
#     """
#     import inspect
#     import os
#     import pickle

#     import h5py
#     import numpy as np
#     import pandas as pd
#     import yaml

#     spath, sfname = None, None

#     if "/" in sfname_or_spath:
#         spath = sfname_or_spath
#     else:
#         sfname = sfname_or_spath

#     if (spath is None) and (sfname is not None):
#         ## for ipython
#         __file__ = inspect.stack()[1].filename
#         if "ipython" in __file__:
#             __file__ = "/tmp/fake.py"

#         ## spath
#         fpath = __file__
#         fdir, fname, _ = split_fpath(fpath)
#         sdir = fdir + fname + "/"
#         spath = sdir + sfname
#         # spath = mk_spath(sfname, makedirs=True)

#     ## Make directory
#     if makedirs:
#         sdir = os.path.dirname(spath)
#         os.makedirs(sdir, exist_ok=True)

#     ## Saves
#     # csv
#     if spath.endswith(".csv"):
#         obj.to_csv(spath)
#     # numpy
#     if spath.endswith(".npy"):
#         np.save(spath, obj)
#     # pkl
#     if spath.endswith(".pkl"):
#         with open(spath, "wb") as s:  # 'w'
#             pickle.dump(obj, s)
#     # png
#     if spath.endswith(".png"):
#         obj.savefig(spath)
#         del obj
#     # tiff
#     if spath.endswith(".tiff") or spath.endswith(".tif"):
#         obj.savefig(spath, dpi=300, format="tiff")  # obj is matplotlib.pyplot object
#         del obj
#     # yaml
#     if spath.endswith(".yaml"):
#         with open(spath, "w") as f:
#             yaml.dump(obj, f)
#     # hdf5
#     if spath.endswith(".hdf5"):
#         name_list, obj_list = []
#         for k, v in obj.items():
#             name_list.append(k)
#             obj_list.append(v)
#         with h5py.File(spath, "w") as hf:
#             for (name, obj) in zip(name_list, obj_list):
#                 hf.create_dataset(name, data=obj)
#     # pth
#     if spath.endswith(".pth"):
#         # torch.save(obj.state_dict(), spath)
#         torch.save(obj, spath)

#     print("\nSaved to: {s}\n".format(s=spath))


# def check_encoding(file_path):
#     from chardet.universaldetector import UniversalDetector

#     detector = UniversalDetector()
#     with open(file_path, mode="rb") as f:
#         for binary in f:
#             detector.feed(binary)
#             if detector.done:
#                 break
#     detector.close()
#     enc = detector.result["encoding"]
#     return enc


# def save_listed_scalars_as_csv(
#     listed_scalars,
#     spath_csv,
#     column_name="_",
#     indi_suffix=None,
#     round=3,
#     overwrite=False,
# ):
#     """Puts to df and save it as csv"""
#     if overwrite == True:
#         mv_to_tmp(spath_csv, L=2)
#     indi_suffix = np.arange(len(listed_scalars)) if indi_suffix is None else indi_suffix
#     df = pd.DataFrame(
#         {"{}".format(column_name): listed_scalars}, index=indi_suffix
#     ).round(round)
#     df.to_csv(spath_csv)
#     print("\nSaved to: {}\n".format(spath_csv))


# def save_listed_dfs_as_csv(listed_dfs, spath_csv, indi_suffix=None, overwrite=False):
#     """listed_dfs:
#         [df1, df2, df3, ..., dfN]. They will be written vertically in the order.

#     spath_csv:
#         /hoge/fuga/foo.csv

#     indi_suffix:
#         At the left top cell on the output csv file, '{}'.format(indi_suffix[i])
#         will be added, where i is the index of the df.On the other hand,
#         when indi_suffix=None is passed, only '{}'.format(i) will be added.
#     """
#     if overwrite == True:
#         mv_to_tmp(spath_csv, L=2)

#     indi_suffix = np.arange(len(listed_dfs)) if indi_suffix is None else indi_suffix
#     for i, df in enumerate(listed_dfs):
#         with open(spath_csv, mode="a") as f:
#             f_writer = csv.writer(f)
#             i_suffix = indi_suffix[i]
#             f_writer.writerow(["{}".format(indi_suffix[i])])
#         df.to_csv(spath_csv, mode="a", index=True, header=True)
#         with open(spath_csv, mode="a") as f:
#             f_writer = csv.writer(f)
#             f_writer.writerow([""])
#     print("Saved to: {}".format(spath_csv))


# def mv_to_tmp(fpath, L=2):
#     import os
#     from shutil import move

#     try:
#         tgt_fname = connect_strs_with_hyphens(fpath.split("/")[-L:])
#         tgt_fpath = "/tmp/{}".format(tgt_fname)
#         move(fpath, tgt_fpath)
#         print("Moved to: {}".format(tgt_fpath))
#     except:
#         pass


# # def makedirs_from_spath(spath):
# #     import os
# #     sdir = os.path.dirname(spath)
# #     os.makedirs(sdir, exist_ok=True)
